#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Basic**

    Basic provides a framework for define, parse, interpret and evaluate
    decorator arguments. It generates and resolves (nested) lists of tuples of
    decorator argument objects between their native forms and F/X structure
    representations.

    Attributes
    ----------
    Activation: IntFlag
        Relative time flags for activity control.

    F: Type
        A wrapper for callee(*callee_args, **callee_kwargs) semantics.

    X: Type(F)
        A cross-wrapper for callee vs. *callee_args and/or **callee_kwargs.

    Parser: Type
        Parser to resolve the function/class decoration argument language.

    Methods
    -------
    None.
"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Basic"
__module__ = "basic.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.4"
__date__ = "2023-06-22"
__time__ = "12:12:42"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["Activation", "Parser", "F", "X"]

# -----------------------------------------------------------------------------
# Libraries & Modules
from enum import IntFlag
from typing import Union


# -----------------------------------------------------------------------------
# Classes
class Activation(IntFlag):
    """Relative time flags for activity control.

    NONE                    No activation
    BEFORE                  Activation before action
    AFTER                   Activation after action
    BEFORE_AND_AFTER        Activation before and after action
    """
    NONE = 0x00
    BEFORE = 0x01
    AFTER = 0x02
    BEFORE_AND_AFTER = BEFORE | AFTER


class F:
    """A wrapper for callee(*callee_args, **callee_kwargs) semantics.

    The caller is designed as a read-only property, since it is used as a key
    to collect F objects in sets.

    The eval() method applies the callee property to its callee_args and/or
    callee_kwargs:
      - If callee is a callable, it is applied directly to the arguments,
      - If callee is a string, eval() is applied to the same-named method of
        the object passed to eval().

    Attributes
    ----------
        callee: callable or str     (getter property)
        callee_args: object
        callee_kwargs: object

    Methods
    -------
    eval(self, obj=None)
        Evaluate the value of callee(callee_args, callee_kwargs)

    Example
    -------

    from decoratory.basic import F

    # Define a callable
    def sandwich(anything="", toast: str = "|"):
        return f"{toast}{anything}{toast}"

    # Case 1: Function argument is a callable
    f = F(sandwich)                 # F(sandwich), sandwich(): ||
    print(f"{repr(f)}, {str(f)}: {f.eval()}")
    f = F(sandwich, 3)              # F(sandwich, 3), sandwich(3): |3|
    print(f"{repr(f)}, {str(f)}: {f.eval()}")
    f = F(sandwich, 'A')            # F(sandwich, "A"), sandwich("A"): |A|
    print(f"{repr(f)}, {str(f)}: {f.eval()}")
    f = F(sandwich, 3, '*')         # F(sandwich, 3, "*"), sandwich(3, "*"):
    print(f"{repr(f)}, {str(f)}: {f.eval()}")   # *3*
    f = F(sandwich, 3, toast='*')   # F(sandwich, 3, toast="*"), sandwich(
    print(f"{repr(f)}, {str(f)}: {f.eval()}")   # 3, toast="*"): *3*

    # Define a method
    class A:
        def sandwich(self, anything="", toast: str = "|"):
            return f"{toast}{anything}{toast}"

    # Case 2: Function argument is a method name
    f = F("sandwich")               # F("sandwich"), sandwich(): ||
    print(f"{repr(f)}, {str(f)}: {f.eval(obj=A())}")
    f = F("sandwich", 3)            # F("sandwich", 3), sandwich(3): |3|
    print(f"{repr(f)}, {str(f)}: {f.eval(obj=A())}")
    f = F("sandwich", 'A')          # F("sandwich", "A"), sandwich("A"): |A|
    print(f"{repr(f)}, {str(f)}: {f.eval(obj=A())}")
    f = F("sandwich", 3, '*')       # F("sandwich", 3, "*"), sandwich(3, "*"):
    print(f"{repr(f)}, {str(f)}: {f.eval(obj=A())}")    # *3*
    f = F("sandwich", 3, toast='*') # F("sandwich", 3, toast="*"), sandwich(
    print(f"{repr(f)}, {str(f)}: {f.eval(obj=A())}")    # 3, toast="*"): *3*
    """
    __slots__ = ('__callee', 'callee_args', 'callee_kwargs')

    @staticmethod
    def __quote(qs: str = "", qq: str = '"') -> str:
        """Special str quotes"""
        return f"{qq}{qs}{qq}" if isinstance(qs, str) else str(qs)

    def __init__(self, callee, *callee_args, **callee_kwargs):
        self.__set__callee(callee)
        self.callee_args = callee_args
        self.callee_kwargs = callee_kwargs

    def __repr__(self):
        s = str(self.__class__.__name__) + '('
        if callable(self.__callee):
            s += str(self.__callee.__name__)
        else:
            s += self.__class__.__quote(self.__callee)
        if self.callee_args:
            s += ', ' + ', '.join(self.__class__.__quote(k)
                                  for k in self.callee_args)
        if self.callee_kwargs:
            s += ', ' + ', '.join((str(k) + '=' + self.__class__.__quote(v))
                                  for k, v in self.callee_kwargs.items())
        return s + ')'

    def __str__(self):
        if callable(self.__callee):
            s = str(self.__callee.__name__) + '('
        else:
            s = str(self.__callee) + '('
        if self.callee_args:
            s += ', '.join(self.__class__.__quote(k)
                           for k in self.callee_args)
        if self.callee_kwargs:
            if self.callee_args:
                s += ', '
            s += ', '.join((str(k) + '=' + self.__class__.__quote(v))
                           for k, v in self.callee_kwargs.items())
        return s + ')'

    def __hash__(self):
        return hash(self.__callee)  # Read only object key, see init above

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.__callee == other.__callee
        elif callable(other) or isinstance(other, str):
            return self.__callee == other
        else:
            return False

    def __iter__(self):
        return iter([self.__callee, self.callee_args, self.callee_kwargs])

    def __next__(self):
        pass  # __iter__ delegates to list iterator

    def eval(self, obj=None):
        """Evaluate the value of callee(callee_args, callee_kwargs)"""
        if callable(self.__callee):
            # Assume default function semantics
            return self.__callee(*self.callee_args, **self.callee_kwargs)
        elif isinstance(self.__callee, str):
            # Assume an instance method semantics
            if obj is None:
                # Assume obj is in first positional argument args[0]
                if self.callee_args:
                    obj = self.callee_args[0]
                    if hasattr(obj, self.__callee):
                        return getattr(obj, self.__callee)(
                            *self.callee_args[1:], **self.callee_kwargs)
            elif hasattr(obj, self.__callee):
                # Assume this obj as the instance of __callee
                return getattr(obj, self.__callee)(
                    *self.callee_args, **self.callee_kwargs)

        # In all other cases
        raise TypeError(f"'{self.__callee}' cannot be evaluated.")

    # Getter, Setter, Properties
    def __get__callee(self):
        return self.__callee

    def __set__callee(self, value: Union[callable, str]):
        self.__callee = value  # Read-only object key, see init above

    callee = property(__get__callee)


class X(F):
    """A cross-wrapper for callee vs. *callee_args and/or **callee_kwargs.

    Technically, X is equivalent to F. Both share the same code, the same
    syntax, but they follow different semantics:
       - F is a wrapper for a callee and its OWN arguments, whereas
       - X is a wrapper for a callee and arguments for another caller

    Although X can be substituted by F, it is recommended to use X to indicate
    that the callee and the specified arguments do not belong to each other,
    for reasons of code clarity.
    """
    pass


class Parser:
    """Parser to resolve the function/class decoration argument language.

    A Parser to resolve the function/class decoration argument language where
    a callable is mandatory and the optional arguments have empty defaults
    into a list of F-structures.

    Syntax:
        Parser.eval(item) -> [F(callable|method_name, *args, **kwarg), ...]

    Production:
        item        :=  list | tuple | F | callable | str | None
        list        := [list | tuple | F | callable | str | None]
            [..,list,..]                -> [..,*list,..]
        tuple       := (..,) | (..,..) | (..,..,..)
            (callable|str,)             -> F(callable|str)
            (callable|str, (..))        -> F(callable|str, *(..))
            (callable|str, {..})        -> F(callable|str, **{..})
            (callable|str, (..), {..})  -> F(callable|str, *(..), **{..})
        F           := F(..) | F(..,..) | F(..,..,..)
            F(callable|str)             -> F(callable|str)
            F(callable|str, *a)         -> F(callable|str, *a)
            F(callable|str, **kwa)      -> F(callable|str, **kwa)
            F(callable|str, *a, **kwa)  -> F(callable|str, *a, **kwa)
        callable    := callable
            callable                    -> F(callable)
        str         := str
            str                         -> F(str)
        None        := None
            None                        -> []
    """

    @classmethod
    def eval(cls, item: Union[list, tuple, F, callable, str, None]) -> list:
        """Format data, resolve and collect as F-structures in a flat list

        list        -> cls.list         (see below)
        tuple       -> [cls.tuple]      (see below)
        F(..,..,..) -> [F(.,.,.)]
        callable    -> [F(callable)]
        string      -> [F(string)]
        None        -> []               (empty list)
        """
        result = list()
        if item is None:
            pass  # Capture the no list default into an empty list
        elif isinstance(item, F):
            result.append(item)
        elif callable(item) or isinstance(item, str):
            result.append(F(item))
        elif isinstance(item, tuple):
            result.append(cls.tuple(item))
        elif isinstance(item, list):
            result.extend(cls.list(item))
        else:
            raise TypeError(f"Object '{item}' cannot be parsed.")
        return result

    @classmethod
    def list(cls, list_obj: list) -> list:
        """Resolve (nested) list recursively.

        Loop over list (iterable) and resolve list items:

        list        -> [..., cls.list]       (extend recursively)
        tuple       -> [..., cls.tuple]      (see below)
        F(..,..,..) -> [..., F(..,..,..)]
        callable    -> [..., F(callable)]
        str         -> [..., F(str)]
        None        -> [...]
        """
        result = list()
        for li in list_obj:
            if isinstance(li, list):
                result.extend(cls.list(li))
            elif isinstance(li, tuple):
                result.append(cls.tuple(li))
            elif isinstance(li, F):
                result.append(li)
            elif callable(li) or isinstance(li, str):
                result.append(F(li))
            elif li is None:
                pass
            else:
                raise TypeError(f"List element '{li}' cannot be parsed.")
        return result

    @classmethod
    def tuple(cls, tuple_obj: tuple) -> F:
        """Resolve tuple and create an F

        Analyse a tuple and map it into an F structure:

        (callable|str,)             -> F(callable|str)
        (callable|str, (..))        -> F(callable|str, *(..))
        (callable|str, {..})        -> F(callable|str, **{..})
        (callable|str, (..), {..})  -> F(callable|str, *(..), **{..})
        """
        if len(tuple_obj) == 1 and \
                (callable(tuple_obj[0]) or isinstance(tuple_obj[0], str)):
            return F(tuple_obj[0])
        elif len(tuple_obj) == 2 and \
                (callable(tuple_obj[0]) or isinstance(tuple_obj[0], str)):
            if isinstance(tuple_obj[1], tuple):
                return F(tuple_obj[0], *tuple_obj[1])
            elif isinstance(tuple_obj[1], dict):
                return F(tuple_obj[0], **tuple_obj[1])
        elif len(tuple_obj) == 3 and \
                (callable(tuple_obj[0]) or isinstance(tuple_obj[0], str)) and \
                isinstance(tuple_obj[1], tuple) and \
                isinstance(tuple_obj[2], dict):
            return F(tuple_obj[0], *tuple_obj[1], **tuple_obj[2])

        # No match
        raise TypeError(f"Tuple '{tuple_obj}' cannot be parsed.")


# -----------------------------------------------------------------------------
# Entry Point
if __name__ == '__main__':
    from decoratory.banner import __banner as banner
    from decoratory import basic

    banner(title=__title__,
           version=__version__,
           date=__date__,
           time=__time__,
           docs=(F, X, Parser, Activation),
           author=__author__,
           maintainer=__maintainer__,
           company=__company__,
           email=__email__,
           url=__url__,
           copyright=__copyright__,
           state=__state__,
           license=__license__)



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Multiton**

    A multiton pattern is a design pattern that extends the singleton pattern.
    Whereas the singleton allows for exactly one instance per class, the
    multiton ensures one single (unique) instance per key.

    In this implementation, the key parameter can be anything that is possible
    as a key for a Python dict() dictionary, such as an immutable type or a
    callable eventually returning such an immutable type etc.

    In case of an invalid key, key is set None and with only one key value the
    multiton simply collapses to a singleton, therefore the decoration
    @Multiton resp. @Multiton() or even @Multiton(key=17) or
    @Multiton(key='some constant value') and so on always creates a singleton.

    Attributes
    ----------
    Multiton (class):
        Creates a multiton instance as a callable object.

    Methods
    -------
        None.

    Examples
    --------

    # -------------------------------------------------------------------------
    from decoratory.multiton import Multiton

    @Multiton(key=lambda spec, name: name)
    class Animal:
        def __init__(self, spec, name):
            self.spec = spec
            self.name = name

        def __repr__(self):
            return f"{self.__class__.__name__}('{self.spec}', '{self.name}')"

    # Create Instances
    a = Animal('dog', name='Teddy')
    b = Animal('cat', name='Molly')
    c = Animal('dog', name='Roxie')

    # Case 0: decoration @Multiton or @Multiton() or @Multiton(key=17) or ...
    #    ---> With no or fixed key the Multiton acts like a Singleton
    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('dog', 'Teddy')
    print(c)                        # Animal('dog', 'Teddy')

    # Case 1: decoration @Multiton(key=lambda spec, name: name)
    #    ---> key is a function evaluating the attribute name from __init__(..)
    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('cat', 'Molly')
    print(c)                        # Animal('dog', 'Roxie')

    # Case 2: decoration @Multiton(key=lambda spec, name: 'a' in name)
    #    ---> key is a function evaluating the attribute name from __init__(..)
    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('cat', 'Molly')
    print(c)                        # Animal('cat', 'Molly')

    # Case 3: decoration @Multiton(key="{0}".format)
    #    ---> Parameter spec is referenced as args[0] (positional)
    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('cat', 'Molly')
    print(c)                        # Animal('dog', 'Teddy')

    # Case 4: decoration @Multiton(key="{name}".format)
    #    ---> Parameter name is referenced as kwargs['name'] (keyword)
    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('cat', 'Molly')
    print(c)                        # Animal('dog', 'Roxie')

    # Case 5: decoration @Multiton(key=lambda spec, name: (spec, name))
    #    ---> One unique instance for all init values, i.e. no duplicates
    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('cat', 'Molly')
    print(c)                        # Animal('dog', 'Roxie')

    # Case 6: decoration @Multiton(key=F("my_key"))
    #    ---> Late binding with F(classmethod_string)
    #         One unique instance from a @staticmethod or @classmethod
    class Animal:
        ...
        @classmethod
        def my_key(cls, spec, name):
            return 'a' in name

    print(a)                        # Animal('dog', 'Teddy')
    print(b)                        # Animal('cat', 'Molly')
    print(c)                        # Animal('cat', 'Molly')

    # Case 7: with decoration @Multiton(key=lambda spec, name: name,
    #                                   resettable=True)
    Animal.reset()                  # Because of resettable=True
    print(Animal.get_instances())   # {}
    print(Animal.issealed())        # False     (=default)
    Animal('dog', name='Teddy')     # Animal('dog', 'Teddy')
    print(Animal.get_instances())   # {'Teddy': Animal('dog', 'Teddy')}
    Animal.seal()                   # Seal the multiton!
    print(Animal.issealed())        # True
    try:                            # Try to..
        Animal('cat', name='Molly') # .. add a new animal
    except  KeyError as ex:         # .. will fail
        print(f"Sorry {ex.args[1]}, {ex.args[0]}")
    print(Animal.get_instances())   # {'Teddy': Animal('dog', 'Teddy')}
    Animal.unseal()                 # Unseal the multiton!
    print(Animal.issealed())        # False
    Animal('cat', name='Molly')     # Now, Molly is added
    print(Animal.get_instances())   # {'Teddy': Animal('dog', 'Teddy'),
                                    #  'Molly': Animal('cat', 'Molly')}
"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Multiton"
__module__ = "multiton.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.2"
__date__ = "2023-06-21"
__time__ = "12:46:42"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["Multiton"]

# -----------------------------------------------------------------------------
# Libraries & Modules
from functools import update_wrapper
from typing import Union
from decoratory.basic import F


# -----------------------------------------------------------------------------
# Classes
class Multiton:
    """**Multiton**

    Multiton(substitute, *args, key, resettable, **kwargs)

    Attributes
    ----------
    substitute (callable|type):
        A type to be made a multiton

    key (F|callable|object|None):
        An object building the instance key

    resettable (bool):
        If True exposes a reset() method

    Methods
    -------
    seal(self) -> None:
        Seal the multiton

    unseal(self) -> None:
        Unseal the multiton

    issealed(self) -> bool:
        Multiton sealing state

    get_instances(self) -> dict:    (property)
        Returns the internal dictionary of instances

    set_instances(self) -> None:    (property)
        Sets the internal dictionary of instances

    reset(self) -> None:            (if resettable=True, only!)
        Resets the multiton instance dictionary
    """

    def __init__(self,
                 substitute: Union[type, callable, None] = None,
                 *args: object,
                 key: Union[F, callable, object, None] = None,
                 resettable: bool = False,
                 **kwargs: object) -> None:
        """Set up a multiton.

        Parameters:
            substitute (object): A type to be made a multiton
            key: (F|callable|object|None): An object building the instance key
            resettable (bool): If True exposes a reset() method

        Returns:
            self (object): Multiton decorator instance
        """
        self.__set__substitute(substitute)
        self.__key = key

        # Dictionary for unique key instances
        self._instances = dict()

        # Sealing state
        self.__sealed = False

        # If resettable == True exposes a reset() method
        if bool(resettable):
            def reset(s: object = self) -> None:
                """Define reset method"""
                s.set_instances(None)

            # Add the reset method
            setattr(self, 'reset', reset)

        # --- Decorator Arguments Pattern (1/2)
        if self.__get__substitute() is not None:
            # Decoration without parameter(s)
            self.__set__substitute(
                F(self.__get__substitute(), *args, **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())

    def __call__(self, *args, **kwargs):
        """Apply the decorator"""
        # --- Decorator Arguments Pattern (2/2)
        if self.__get__substitute() is None:
            # Decoration with parameter(s)
            self.__set__substitute(F(args[0], *args[1:], **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())
            return self
        else:  # *** Wrapper ***
            # If no current values, take defaults
            if args or kwargs:
                subst = F(self.__get__substitute().callee, *args, **kwargs)
            else:
                subst = self.__get__substitute()

            # Calculate key from callable or read key from arguments
            try:
                if isinstance(self.__key, F):  # classmethod or staticmethod
                    if callable(self.__key.callee):
                        d_key = F(self.__key.callee, *subst.callee_args,
                                  **subst.callee_kwargs).eval()
                    else:
                        d_key = F(
                            getattr(self.substitute.callee, self.__key.callee),
                            *subst.callee_args, **subst.callee_kwargs).eval()
                elif callable(self.__key):  # function
                    d_key = F(self.__key, *subst.callee_args,
                              **subst.callee_kwargs).eval()
                else:  # Value
                    d_key = self.__key
                instance = self._instances.get(d_key, None)
            except (TypeError, Exception):  # Default is None
                d_key = None
                instance = self._instances.get(d_key, None)

            # Create and store new or return existing instance (by key)
            if instance is None:
                if self.__sealed:
                    raise KeyError(f"{self.__name__} is sealed.", d_key)
                instance = self._instances.setdefault(d_key, subst.eval())

            return instance

    # Getter, Setter, Properties
    def __get__substitute(self):
        return self.__substitute

    def __set__substitute(self, value):
        self.__substitute = value

    substitute = property(__get__substitute)

    def get_instances(self) -> dict:
        """Returns the internal dictionary of instances"""
        return self._instances

    def set_instances(self, value: dict = None) -> None:
        """Sets the internal dictionary of instances"""
        self._instances = value if isinstance(value, dict) else dict()

    instances = property(get_instances, set_instances)

    # Methods
    def seal(self) -> None:
        """Seal multiton.

        Parameters:
            None.

        Returns:
            None.
        """
        self.__sealed = True

    def unseal(self) -> None:
        """Unseal multiton .

        Parameters:
            None.

        Returns:
            None.
        """
        self.__sealed = False

    def issealed(self) -> bool:
        """Multiton sealing state

        Parameters:
            None.

        Returns:
            True/False (bool): Sealing state.
        """
        return self.__sealed


# -----------------------------------------------------------------------------
# Entry Point
if __name__ == '__main__':
    import decoratory.multiton as module
    from decoratory.banner import __banner as banner

    banner(title=__title__,
           version=__version__,
           date=__date__,
           time=__time__,
           docs=(module, Multiton),
           author=__author__,
           maintainer=__maintainer__,
           company=__company__,
           email=__email__,
           url=__url__,
           copyright=__copyright__,
           state=__state__,
           license=__license__)



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Decoratory**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Decoratory"
__module__ = "__main__.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.6"
__date__ = "2023-06-23"
__time__ = "09:10:15"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["get_file_location"]

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest
import sys

from os.path import dirname, join, basename
from glob import glob
from site import getsitepackages, getusersitepackages
from importlib import import_module

# -----------------------------------------------------------------------------
# Parameters
module = __title__.lower()
name = __title__.capitalize()


# -----------------------------------------------------------------------------
# Functions
def get_file_location(filename: str = "License.txt"):
    """Finds an installed data file in any OS dependent installation"""
    # Look in all possible site package locations, global and user
    if isinstance(getsitepackages(), list):
        package_locations = getsitepackages()
    else:
        package_locations = [getsitepackages()]
    if isinstance(getusersitepackages(), list):
        package_locations.extend(getusersitepackages())
    else:
        package_locations.append(getusersitepackages())
    # Try to find the RECORD file in dist/version folder in package_location
    target_path = None
    for package_location in package_locations:
        try:
            # Read RECORD file
            record = join(package_location,
                          f"{__title__.lower()}-{__version__}.dist-info",
                          "RECORD")
            with open(record, "r") as f_rec:
                rec_lines = f_rec.readlines()
            # Search in lines for the given filename
            for rec_line in rec_lines:
                file_path, sep, *skip = rec_line.partition(",")
                if file_path and str(file_path).endswith(filename):
                    try:
                        path = join(package_location, file_path)
                        with open(path, "r") as fp:
                            fp.read(1)  # something to read...?
                        target_path = path
                        break
                    except (FileNotFoundError, Exception):
                        continue
        except (FileNotFoundError, Exception):
            continue
    return target_path


# -----------------------------------------------------------------------------
# A Module Entry Point
def main():
    """Package information"""
    arg = sys.argv[1].strip().lower() if len(sys.argv) > 1 else "message"
    if arg in "--version":
        print(f"{module}: Version {__version__}, "
              f"Build {__date__} {__time__}, State '{__state__}'")
    elif arg in "--copyright":
        print(f"{__copyright__}")
    elif arg in "--license" or arg in "--licence":
        print(f"\n{'-' * 79}")
        print(f"{__title__} ships under the license: {__license__}".center(79))
        print(f"{'-' * 79}")
        file_path = get_file_location("License.txt")
        if file_path is not None:
            with open(file_path, "r") as fp:
                txt = fp.read()
            print("\n" + txt)
        else:
            print("*** No additional data accessible. Please look for "
                  "file 'License.txt' ***".center(79, " "))
        # print(f"{'-' * 79}")
        # print(f"File: {file_path}".center(79, " "))
        print(f"{'-' * 79}")
    elif arg in "--test":
        try:
            print(f"\n{'-' * 70}")
            print(f" Running unit tests for all modules ".center(70, '-'))
            print(f"{'-' * 70}\n")
            loader = unittest.TestLoader().loadTestsFromModule
            sys_path = dirname(get_file_location(f"test_basic.py"))
            if sys_path and sys_path not in sys.path:
                sys.path.insert(0, sys_path)
            for mdl in glob(join(sys_path, "test_*.py")):
                mdl = basename(mdl)[:-3].lower()
                print(f"\n{' ' + mdl + ' ' :-^70s}\n")
                mdl = import_module(mdl)
                suite = loader(mdl)
                unittest.TextTestRunner(verbosity=2).run(suite)
        except (ModuleNotFoundError, ImportError, NameError, Exception):
            print(f"*** No unit test available for package "
                  f"{module} ***\n".center(70, " "))
        print(f"{'-' * 70}")
    elif arg in "--information":
        print(f"\nPeople and contact information for package {module}:\n")
        print(f"Author    : {__author__}")
        print(f"Maintainer: {__maintainer__}")
        print(f"Company   : {__company__}")
        print(f"Email     : {__email__}")
        print(f"Web       : {__url__}")
    else:
        print(f"""
-------------------------------------------------------------------------------
{__title__.upper(): ^79}
-------------------------------------------------------------------------------

{module} is a Python package you cannot execute directly from command line.

The package provides a framework for Python decorators as well as concrete 
implementations of useful decorators, e.g. Singleton, Multiton, Observable 
and a configurable Wrapper.

{module} is provided by {__company__} under the {__license__} license \
included with 
this package. It comes without any warranty, but with the intention of saving 
working time and improving code quality and productivity.

All kinds of bugs, questions, improvement suggestions, change requests etc. 
are welcome to be directed to the product maintainers email {__email__}.

The current version of the installed package is:
{module}: Version {__version__}, Build {__date__} {__time__}, \
State '{__state__}'

Syntax:   python -m {module}[.module] [option]

For information about the whole package the module command can be left empty, 
for special module information one of the following modules can be addressed:

{'singleton, multiton, wrapper, observer': ^79s}

Possible options are:

  -h, --help           show this help message
  -v, --version        show {module}[.module] version information      
  -c, --copyright      show {module}[.module] copyright statement
  -l, --license        show {module}[.module] license statement (License.txt)
      --licence        an alias to --license
  -t, --test           run unit tests for {module}[.module]
  -i, --info           people and contact information

Example: Display license statement for the multiton module:

{'python -m decoratory.multiton --license': ^79s}
         
-------------------------------------------------------------------------------
{__copyright__: <69s}{__date__: >10s}
-------------------------------------------------------------------------------
""")


# -----------------------------------------------------------------------------
# Execution
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
"""**Decorators**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "decoratory"
__module__ = "setup.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.6"
__date__ = "2023-06-23"
__time__ = "09:10:15"
__state__ = "Beta"
__license__ = "PSF"

# -----------------------------------------------------------------------------
# Libraries
from os.path import join
from setuptools import setup, find_packages
from setuptools.command.install import install

# -----------------------------------------------------------------------------
# Parameters
src = "Sources"
dta = join("lib", "site-packages", __title__, "data")
tst = join("lib", "site-packages", __title__, "tests")

with open("Readme.rst", "r") as f:
    description = f.read()

with open("Requirements.txt", "r") as f:
    requirements = [str(req) for req in f.read().splitlines() if req]


def version_check():
    """Keep versions in sync"""
    try:
        # noinspection PyProtectedMember
        from Sources.decoratory.__main__ import __version__ as version
        assert version == __version__, (
            f"\n\n{'':=^79s}\n"
            f"{' Version problem: __main__.version != setup.version ':=^79s}\n"
            f"{'':=^79s}")
    except (ModuleNotFoundError, ImportError):
        pass  # Let the AssertionError pass...


# -----------------------------------------------------------------------------
# Excecute
setupargs = dict(
    # General
    name=__title__,
    version=__version__,
    author=__author__,
    author_email=f'{__author__} <{__email__}>',
    maintainer=f'{__author__}',
    maintainer_email=f'{__author__} <{__email__}>',
    url=__url__,
    download_url=__url__,
    description=('Decorators: Singleton, Multiton, Observer, Observable, '
                 'generic Wrapper.'),
    long_description=description,
    long_description_content_type='text/x-rst',
    project_urls={
        'Projekt': __url__,
        'Release Notes': __url__,
        'Download': __url__},
    keywords='decorator singleton multiton observer observable wrapper',
    # Technical
    license=__license__,
    platforms=['Operating System :: OS Independent'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation',
        'License :: OSI Approved :: Python Software Foundation License',
        'Operating System :: OS Independent',
        'Topic :: Education',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities'],
    # Modules, Files and Data
    # https://setuptools.pypa.io/en/latest/userguide/package_discovery.html#finding-simple-packages
    packages=find_packages(where=src),
    package_dir={"": src},
    package_data={},
    py_modules=[],
    data_files=[(dta, ["License.txt",
                       "Readme.rst",
                       "Requirements.txt"]),
                (tst, ["Unittest/test_basic.py",
                       "Unittest/test_singleton.py",
                       "Unittest/test_multiton.py",
                       "Unittest/test_wrapper.py",
                       "Unittest/test_observer.py"])],
    entry_points={},
    # Dependencies
    python_requires='>=3.7',
    setup_requires=[],
    install_requires=requirements,
    # Post install
    # cmdclass={'install': CustomInstall},
)
setup(**setupargs)

version_check()



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Singleton**

    A singleton pattern is a design pattern that limits the instantiation of a
    class to a single (unique) instance. This is useful when exactly one
    unique object is needed i.e. to manage an expensive resource or coordinate
    actions across module boundaries.

    Attributes
    ----------
    Singleton (class):
        Creates a singleton instance as a callable object.

    SemiSingleton (class):
        Creates a resettable semi singleton instance as a callable object.

    Methods
    -------
        None.

    Examples
    --------

    from decoratory.singleton import Singleton

    # -------------------------------------------------------------------------
    @Singleton                      # or @Singleton()
    class Animal:
        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return f"{self.__class__.__name__}('{self.name}')"

    # Create Instances
    a = Animal(name='Teddy')        # Creates Teddy, the primary instance
    b = Animal(name='Roxie')        # Returns Teddy, no Roxi is created

    # Case 1: Static decoration using @Singleton or @Singleton()
    print(f"a = {a}")               # a = Animal('Teddy')
    print(f"b = {b}")               # b = Animal('Teddy')
    print(f"a is b: {a is b}")      # a is b: True
    print(f"a == b: {a == b}")      # a == b: True

    # Case 2: Dynamic decoration providing extra initial default values
    Animal = Singleton(Animal, 'Teddy')
    Animal()                        # Using the decorator's default 'Teddy'
    a = Animal(name='Roxie')        # Returns Teddy
    print(a)                        # Animal('Teddy')

    # -------------------------------------------------------------------------
    @Singleton(resettable=True)     # Exposes an additional reset method
    class Animal:
        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return f"{self.__class__.__name__}('{self.name}')"

    # Case 3: Decoration using @Singleton(resettable=True)
    print(Animal(name='Teddy'))     # Animal('Teddy')
    print(Animal(name='Roxie'))     # Animal('Teddy')   (=primary instance)
    Animal.reset()                  # Reset the singleton
    print(Animal(name='Roxie'))     # Animal('Roxie')
    print(Animal(name='Teddy'))     # Animal('Roxie')   (=primary instance)

    # -------------------------------------------------------------------------
    from decoratory.singleton import SemiSingleton

    @SemiSingleton                  # or @SemiSingleton()
    class Animal:
        pass                        # Some code ...
"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Singleton"
__module__ = "singleton.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.2"
__date__ = "2023-06-21"
__time__ = "12:46:42"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["Singleton", "SemiSingleton"]

# -----------------------------------------------------------------------------
# Libraries & Modules
from functools import update_wrapper
from typing import Union
from decoratory.basic import F


# -----------------------------------------------------------------------------
# Classes
class Singleton:
    """**Singleton**

    Singleton(substitute, *args, resettable, **kwargs)

    Attributes
    ----------
    substitute (callable|type):
        A type to be made a singleton

    resettable (bool):
        If True exposes a reset() method

    Methods
    -------
    get_instance(self) -> object:
        Return the singleton instance (primary instance)

    reset(self) -> None:            (if resettable=True, only!)
        Resets the singleton instance (None)
    """

    def __init__(self,
                 substitute: Union[type, callable, None] = None,
                 *args: object,
                 resettable: bool = False,
                 **kwargs: object) -> None:
        """Set up a singleton.

        Parameters:
            substitute (object): A type to be made a singleton
            resettable (bool): If True exposes a reset() method

        Returns:
            self (object): Singleton decorator instance
        """
        self.__set__substitute(substitute)

        # The unique instance
        self.__instance = None

        # If resettable == True exposes a reset() method
        if bool(resettable):
            def reset(s: object = self) -> None:
                """Define reset method"""
                s.__instance = None

            # Add the reset method
            setattr(self, 'reset', reset)

        # --- Decorator Arguments Pattern (1/2)
        if self.__get__substitute() is not None:
            # Decoration without parameter(s)
            self.__set__substitute(
                F(self.__get__substitute(), *args, **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())

    def __call__(self, *args, **kwargs):
        """Apply the decorator."""

        # --- Decorator Arguments Pattern (2/2)
        if self.__get__substitute() is None:
            # Decoration with parameter(s)
            self.__set__substitute(F(args[0], *args[1:], **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())
            return self
        else:  # *** Wrapper ***
            # Create and store new or return existing instance
            if self.__instance is None:
                if args or kwargs:
                    self.__instance = F(self.__get__substitute().callee,
                                        *args, **kwargs).eval()
                else:
                    self.__instance = self.__get__substitute().eval()
            return self.__instance

    # Getter, Setter, Properties
    def __get__substitute(self):
        return self.__substitute

    def __set__substitute(self, value):
        self.__substitute = value

    substitute = property(__get__substitute)

    # Methods
    def get_instance(self) -> object:
        """Returns the singleton instance"""
        return self.__instance


class SemiSingleton(Singleton):
    """SemiSingleton

    A subclass shortcut for a resettable singleton.
    """

    def __init__(self,
                 substitute: object = None,
                 *args: object,
                 **kwargs: object) -> None:
        """Set up a semi singleton.

        Parameters:
            substitute (object): A type to be made a semi singleton

        Returns:
            self (object): Semi singleton decorator instance
        """
        kwargs["resettable"] = True
        super().__init__(substitute, *args, **kwargs)


# -----------------------------------------------------------------------------
# Entry Point
if __name__ == '__main__':
    import decoratory.singleton as module
    from decoratory.banner import __banner as banner

    banner(title=__title__,
           version=__version__,
           date=__date__,
           time=__time__,
           docs=(module, Singleton),
           author=__author__,
           maintainer=__maintainer__,
           company=__company__,
           email=__email__,
           url=__url__,
           copyright=__copyright__,
           state=__state__,
           license=__license__)
